/*
 *
 * gnumbers.h
 *
 */

#ifndef GNUMBERS_H
#define GNUMBERS_H

#include <iostream.h>
#include <math.h>

//
// These classes are used to represent GMatrices, GPoints and GVectors.
// They have assembly-like operations. 
// mld = multiply - load
// mac = multiply - accumulate
// msb = multiply - substract
// lod = load
// ldo = load opposite ( = lod(-x) )
// sto = store
// mul = multiply
// div = divide
//
// For the moment, they are floats, but it will be integers later on ...
// They make matrix-vector computation independant from the internal
// representation.
//
// I could have overloaded the operators =, *=, += ..., but I want to
// be sure about what conversions are done, and there is no operator
// for mac, mld, and msb ...
//

typedef float gfloat;
typedef int   gint;
typedef int   GAngle;

// Coefficient of a matrix.

class GCoeff
{
 public:

  void mld(GCoeff x1, GCoeff x2);
  void mac(GCoeff x1, GCoeff x2);
  
  void lod(GCoeff x);
  void ldo(GCoeff x);
  void lod(gint   x);
  void lod(gfloat x);

  void sto(GCoeff &x);
  void sto(gint   &x);
  void sto(gfloat &x);

  gint   iget(void);
  gfloat fget(void); 

  void   sin(GAngle alpha); // angles are in tenth of degrees
  void   cos(GAngle alpha);

  gfloat& data(void);

 private:
  gfloat _x;

  friend class GCoord;
};

// ostream& operator<<(ostream& output, GCoeff x);

// Coordinate

class GCoord
{
 public:
  void mld(GCoord x1, GCoeff x2);
  void mac(GCoord x1, GCoeff x2);
  
  void mld(GCoord x1, GCoord x2);
  void mac(GCoord x1, GCoord x2);
  void msb(GCoord x1, GCoord x2);

  void lod(GCoord x);
  void lod(gint   x);
  void lod(gfloat x);

  void sto(GCoord &x);
  void sto(gint   &x);
  void sto(gfloat &x);

  void div(GCoord x);

  gint   iget(void);
  gfloat fget(void); 

  gfloat& data(void);

 private:
  gfloat _x;
};

// ostream& operator<<(ostream& output, GCoord x);

#include "gnum.ih"

#endif

