/*
 *
 * polyeng.h
 *
 */


#ifndef POLYENG_H
#define POLYENG_H

#include "gdefs.h"
#include "vpool.h"
#include "polygon.h"
#include "gproc.h"

const int PE_MAX_HEIGHT = 1024;

const int D_SHIFT  = 4;
const int D_MASK   = 15;
const int D_PMASK  = 3;
const int D_PSHIFT = 2;

const int PEA_ZCLIP = GA_MAX + 1;

typedef struct 
{
  GVertex _left;
  GVertex _right;
} Mug;


class PolygonEngine;

typedef void (* PolyFun)(GVertexAttributes *);
typedef void (* LineFun)(GVertexAttributes *, GVertex *, GVertex *);
typedef void (* PixelFun)(GVertexAttributes *, ScrCoord, ScrCoord);


typedef PolygonEngine *(*PolygonEngine_MakeFun)(GraphicPort*, int);

typedef struct
{
  PolygonEngine_MakeFun make;
  int bytes_per_pixel;
  int R_mask;
  int G_mask;
  int B_mask;
} PolygonEngine_VC_Entry;


class PolygonEngine : public GraphicProcessor
{

 protected:
  GVertexAttributes    _va;

  static GVertexPool _vpool;
  static GPolygon    _p1;
  static GPolygon    _p2;
  static Mug         _mug[PE_MAX_HEIGHT];

  static   int D4[16]; // Dithering matrix

  GVertexAttributes _vattrib_stack[ATTRIB_STACK_SZ];

  PolyFun   _fillpoly;
  LineFun   _drawline;
  PixelFun  _setpixel;

 public:
  PolygonEngine(GraphicPort *gp, int verbose_level = MSG_NONE);
  virtual ~PolygonEngine(void);
  virtual void CommitAttributes(void) = 0;
  GVertexAttributes& VAttributes(void);

  void Push(GVertex *v);
  void Reset(void);

  void FillPoly();
  void DrawPoly();

  void SetPixel(ScrCoord x, ScrCoord y);
  void SetPixel(ScrCoord x, ScrCoord y, ColorIndex c);
  void SetPixel(ScrCoord x, ScrCoord y, 
		ColorComponent r, ColorComponent g, ColorComponent b);

  void ClipPoly(void);

  // virtual constructor stuff

 public:
  static PolygonEngine* Make(GraphicPort *GP, int verb = 0);

 private:
  static int _entries;
  static PolygonEngine_VC_Entry _entry[VC_ENTRIES];
  static void Register(const PolygonEngine_MakeFun make, int bytes_per_pixel,
		int R_mask, int G_mask, int B_mask);

  friend class Stub;
};

#include "polyeng.ih"

#endif
