/*
 *
 * gport.C
 *
 */

#include "gport.h"
#include <stdio.h>
#include <string.h>

// A virtual destructor for GraphicPorts

GraphicPort::~GraphicPort(void)
{
}


// RGB to ColorCode translation using graphic endianness

UColorCode GraphicPort::RGB2ColorCode(ColorComponent R, 
				      ColorComponent G,
				      ColorComponent B)
{
int r_shift = firstbit(_R_mask);
int g_shift = firstbit(_G_mask);
int b_shift = firstbit(_B_mask);

int r_shift_2 = 8 - nbrbits(_R_mask);
int g_shift_2 = 8 - nbrbits(_G_mask);
int b_shift_2 = 8 - nbrbits(_B_mask);

return   ((((UColorCode)R) >> r_shift_2) << r_shift) |
         ((((UColorCode)G) >> g_shift_2) << g_shift) |
	 ((((UColorCode)B) >> b_shift_2) << b_shift);

}

// Virtual constructor

int GraphicPort::_entries = 0;
GraphicPort_VC_Entry GraphicPort::_entry[VC_ENTRIES];  

void GraphicPort::Register(const GraphicPort_MakeFun make, const int flags)
{
  assert(_entries < VC_ENTRIES);
  _entry[_entries].make    = make;
  _entry[_entries].flags   = flags;
  _entries++;

#ifdef EBUG
  cerr << "[TAGL_VC] GraphicPort::register() called - " << _entries << " entries\n";
#endif
}

GraphicPort* GraphicPort::Make(char* name, ScrCoord width, ScrCoord height, 
				   int verb)
{
  int i;
  GraphicPort *GP = NULL;
  for(i=0; i<_entries; i++)
    if((!_entry[i].flags) && (GP = _entry[i].make(name, width, height, verb)))
      return GP;

  for(i=0; i<_entries; i++)
    if((_entry[i].flags) && (GP = _entry[i].make(name, width, height, verb)))
      return GP;

#ifdef EBUG
  cerr << "[TAGL_VC] GraphicPort_VCC:: No constructor available for this technology\n";
#endif
  return NULL;
}


void GraphicPort::Clear(UColorCode x)
{
  typedef struct
    {
      unsigned char r,g,b;
    } Pixel24;

  switch(_bytes_per_pixel)
    {
    case 1:
      memset(_graph_mem, (unsigned char)x, _bytes_per_line * _height);
      break;
    case 2:
      {
	unsigned short *graph_ptr;
	int size = (_bytes_per_line * _height) >> 1;
	for(graph_ptr = (unsigned short *)_graph_mem; size > 0; size--)
	  *(graph_ptr++) = (unsigned short)x;
      }
      break;
    case 3:
      {
	Pixel24 *graph_ptr;
	int size = (_bytes_per_line * _height) / 3;
	for(graph_ptr = (Pixel24 *)_graph_mem; size > 0; size--)
	  *(graph_ptr++) = *(Pixel24 *)&x; // little endian only !! :-(
      }
      break;
    case 4:
      {
	unsigned long *graph_ptr;
	int size = (_bytes_per_line * _height) >> 2;
	for(graph_ptr = (unsigned long *)_graph_mem; size > 0; size--)
	  *(graph_ptr++) = (unsigned long)x;
      }
    default:
      (*this)[MSG_ERROR]   << "No code to clear this graphic port\n";
      (*this)[MSG_WARNING] << "Setting pixel values to 0\n";
      memset(_graph_mem, 0, _bytes_per_line * _height);
      break;
    }
}

void GraphicPort::ZClear(void)
{
  memset(_z_mem, 255, _width * _height * sizeof(ZCoord));
}

void GraphicPort::ZClear(ZCoord z)
{
  ZCoord *z_ptr;
  int size = _width * _height;
  for(z_ptr = _z_mem; size > 0; size--)
    *(z_ptr++) = z;
}
